require 'httparty'
require 'digest/md5'

require 'galactus/response'

module Galactus
  class Client
    include HTTParty

    class InvalidClientError < StandardError; end

    attr_accessor :api_version, :router
    attr_reader :public_api_key, :private_api_key

    base_uri 'http://gateway.marvel.com'

    def initialize(settings)
      validate_client(settings)
      @public_api_key = settings[:public_api_key]
      @private_api_key = settings[:private_api_key]
      @api_version = settings[:api_version] || 'v1'
      @router = settings[:router] || Galactus.router(api_version: api_version)
      provide_endpoint_methods
    end

    private

    def validate_client(settings)
      [:public_api_key, :private_api_key].each do |param|
        unless settings[param]
          raise InvalidClientError, "A :#{param} parameter must be specified."
        end
      end
    end

    def provide_endpoint_methods
      router.routes.each do |_route_path, route_info|
        name = route_info[:name]
        self.class.send(:define_method, name) do |*args|
          params, headers = process_parameters(*args)
          response = fetch_response(name, params, headers)
          Response.create(response)
        end
      end
    end

    def process_parameters(*args)
      id_holder, options_holder = *args

      if id_holder && id_holder.is_a?(Hash)
        id = nil
        options = id_holder || {}
      elsif id_holder
        id = id_holder
        options = options_holder || {}
      else
        options = {}
      end

      params, headers = pull_headers(options)

      [{ id: id, options: params }, headers]
    end

    def fetch_response(name, params = {}, headers = {})
      id = params[:id]
      options = params[:options]
      path = find_path(name, id)

      self.class.get(path, query: marvel_api_params(options), headers: headers)
    end

    def fetch_resource_id(name, id)
      endpoint = name.split('_')[0] + 's'
      resource = find_by_name_or_title(endpoint.to_sym, :name, id)
      return false unless resource
      resource[:id]
    end

    def pull_headers(options)
      headers = options.delete(:headers) || {}
      unless headers['Accept-Encoding']
        headers.merge!('Accept-Encoding' => 'gzip')
      end
      [options, headers]
    end

    def find_by_name_or_title(endpoint, column, value)
      response = send(endpoint, column.to_sym => value)
      return false unless response[:data][:count] > 0
      response[:id] = response[:data][:results][0][:id]
      response
    end

    def find_path(name, id)
      if id.nil?
        router.send("#{name}_path".to_sym)
      else
        id = fetch_resource_id(name, id) if id.is_a?(String)
        router.send("#{name}_path".to_sym, id: id)
      end
    end

    def marvel_api_params(params = {})
      base_hash = {
        apikey: public_api_key,
        ts: timestamp,
        hash: api_request_hash }

      params.merge(base_hash)
    end

    def timestamp
      Time.now.to_i
    end

    def api_request_hash
      Digest::MD5.hexdigest("#{timestamp}#{private_api_key}#{public_api_key}")
    end
  end
end
